from __future__ import print_function, unicode_literals, absolute_import, division

import numpy as np
import warnings
import os
import datetime
from tqdm import tqdm
from collections import defaultdict
from zipfile import ZipFile, ZIP_DEFLATED
from scipy.ndimage.morphology import distance_transform_edt, binary_fill_holes
from scipy.ndimage.measurements import find_objects
from scipy.optimize import minimize_scalar
from skimage.measure import regionprops
from csbdeep.utils import _raise
from csbdeep.utils.six import Path
from collections.abc import Iterable

from .matching import matching_dataset, _check_label_array


try:
    from edt import edt
    _edt_available = True
    try:    _edt_parallel_max = len(os.sched_getaffinity(0))
    except: _edt_parallel_max = 128
    _edt_parallel_default = 4
    _edt_parallel = os.environ.get('STARDIST_EDT_NUM_THREADS', _edt_parallel_default)
    try:
        _edt_parallel = min(_edt_parallel_max, int(_edt_parallel))
    except ValueError as e:
        warnings.warn(f"Invalid value ({_edt_parallel}) for STARDIST_EDT_NUM_THREADS. Using default value ({_edt_parallel_default}) instead.")
        _edt_parallel = _edt_parallel_default
    del _edt_parallel_default, _edt_parallel_max
except ImportError:
    _edt_available = False
    # warnings.warn("Could not find package edt... \nConsider installing it with \n  pip install edt\nto improve training data generation performance.")
    pass


def gputools_available():
    try:
        import gputools
    except:
        return False
    return True


def path_absolute(path_relative):
    """ Get absolute path to resource"""
    base_path = os.path.abspath(os.path.dirname(__file__))
    return os.path.join(base_path, path_relative)


def _is_power_of_2(i):
    assert i > 0
    e = np.log2(i)
    return e == int(e)


def _normalize_grid(grid,n):
    try:
        grid = tuple(grid)
        (len(grid) == n and
         all(map(np.isscalar,grid)) and
         all(map(_is_power_of_2,grid))) or _raise(TypeError())
        return tuple(int(g) for g in grid)
    except (TypeError, AssertionError):
        raise ValueError("grid = {grid} must be a list/tuple of length {n} with values that are power of 2".format(grid=grid, n=n))


def edt_prob(lbl_img, anisotropy=None):
    if _edt_available:
        return _edt_prob_edt(lbl_img, anisotropy=anisotropy)
    else:
        # warnings.warn("Could not find package edt... \nConsider installing it with \n  pip install edt\nto improve training data generation performance.")
        return _edt_prob_scipy(lbl_img, anisotropy=anisotropy)

def _edt_prob_edt(lbl_img, anisotropy=None):
    """Perform EDT on each labeled object and normalize.
    Internally uses https://github.com/seung-lab/euclidean-distance-transform-3d
    that can handle multiple labels at once
    """
    lbl_img = np.ascontiguousarray(lbl_img)
    constant_img = lbl_img.min() == lbl_img.max() and lbl_img.flat[0] > 0
    if constant_img:
        warnings.warn("EDT of constant label image is ill-defined. (Assuming background around it.)")
    # we just need to compute the edt once but then normalize it for each object
    prob = edt(lbl_img, anisotropy=anisotropy, black_border=constant_img, parallel=_edt_parallel)
    objects = find_objects(lbl_img)
    for i,sl in enumerate(objects,1):
        # i: object label id, sl: slices of object in lbl_img
        if sl is None: continue
        _mask = lbl_img[sl]==i
        # normalize it
        prob[sl][_mask] /= np.max(prob[sl][_mask]+1e-10)
    return prob

def _edt_prob_scipy(lbl_img, anisotropy=None):
    """Perform EDT on each labeled object and normalize."""
    def grow(sl,interior):
        return tuple(slice(s.start-int(w[0]),s.stop+int(w[1])) for s,w in zip(sl,interior))
    def shrink(interior):
        return tuple(slice(int(w[0]),(-1 if w[1] else None)) for w in interior)
    constant_img = lbl_img.min() == lbl_img.max() and lbl_img.flat[0] > 0
    if constant_img:
        lbl_img = np.pad(lbl_img, ((1,1),)*lbl_img.ndim, mode='constant')
        warnings.warn("EDT of constant label image is ill-defined. (Assuming background around it.)")
    objects = find_objects(lbl_img)
    prob = np.zeros(lbl_img.shape,np.float32)
    for i,sl in enumerate(objects,1):
        # i: object label id, sl: slices of object in lbl_img
        if sl is None: continue
        interior = [(s.start>0,s.stop<sz) for s,sz in zip(sl,lbl_img.shape)]
        # 1. grow object slice by 1 for all interior object bounding boxes
        # 2. perform (correct) EDT for object with label id i
        # 3. extract EDT for object of original slice and normalize
        # 4. store edt for object only for pixels of given label id i
        shrink_slice = shrink(interior)
        grown_mask = lbl_img[grow(sl,interior)]==i
        mask = grown_mask[shrink_slice]
        edt = distance_transform_edt(grown_mask, sampling=anisotropy)[shrink_slice][mask]
        prob[sl][mask] = edt/(np.max(edt)+1e-10)
    if constant_img:
        prob = prob[(slice(1,-1),)*lbl_img.ndim].copy()
    return prob


def _fill_label_holes(lbl_img, **kwargs):
    lbl_img_filled = np.zeros_like(lbl_img)
    for l in (set(np.unique(lbl_img)) - set([0])):
        mask = lbl_img==l
        mask_filled = binary_fill_holes(mask,**kwargs)
        lbl_img_filled[mask_filled] = l
    return lbl_img_filled


def fill_label_holes(lbl_img, **kwargs):
    """Fill small holes in label image."""
    # TODO: refactor 'fill_label_holes' and 'edt_prob' to share code
    def grow(sl,interior):
        return tuple(slice(s.start-int(w[0]),s.stop+int(w[1])) for s,w in zip(sl,interior))
    def shrink(interior):
        return tuple(slice(int(w[0]),(-1 if w[1] else None)) for w in interior)
    objects = find_objects(lbl_img)
    lbl_img_filled = np.zeros_like(lbl_img)
    for i,sl in enumerate(objects,1):
        if sl is None: continue
        interior = [(s.start>0,s.stop<sz) for s,sz in zip(sl,lbl_img.shape)]
        shrink_slice = shrink(interior)
        grown_mask = lbl_img[grow(sl,interior)]==i
        mask_filled = binary_fill_holes(grown_mask,**kwargs)[shrink_slice]
        lbl_img_filled[sl][mask_filled] = i
    return lbl_img_filled


def sample_points(n_samples, mask, prob=None, b=2):
    """sample points to draw some of the associated polygons"""
    if b is not None and b > 0:
        # ignore image boundary, since predictions may not be reliable
        mask_b = np.zeros_like(mask)
        mask_b[b:-b,b:-b] = True
    else:
        mask_b = True

    points = np.nonzero(mask & mask_b)

    if prob is not None:
        # weighted sampling via prob
        w = prob[points[0],points[1]].astype(np.float64)
        w /= np.sum(w)
        ind = np.random.choice(len(points[0]), n_samples, replace=True, p=w)
    else:
        ind = np.random.choice(len(points[0]), n_samples, replace=True)

    points = points[0][ind], points[1][ind]
    points = np.stack(points,axis=-1)
    return points


def calculate_extents(lbl, func=np.median):
    """ Aggregate bounding box sizes of objects in label images. """
    if (isinstance(lbl,np.ndarray) and lbl.ndim==4) or (not isinstance(lbl,np.ndarray) and  isinstance(lbl,Iterable)):
        return func(np.stack([calculate_extents(_lbl,func) for _lbl in lbl], axis=0), axis=0)

    n = lbl.ndim
    n in (2,3) or _raise(ValueError("label image should be 2- or 3-dimensional (or pass a list of these)"))

    regs = regionprops(lbl)
    if len(regs) == 0:
        return np.zeros(n)
    else:
        extents = np.array([np.array(r.bbox[n:])-np.array(r.bbox[:n]) for r in regs])
        return func(extents, axis=0)


def polyroi_bytearray(x,y,pos=None,subpixel=True):
    """ Byte array of polygon roi with provided x and y coordinates
        See https://github.com/imagej/imagej1/blob/master/ij/io/RoiDecoder.java
    """
    import struct
    def _int16(x):
        return int(x).to_bytes(2, byteorder='big', signed=True)
    def _uint16(x):
        return int(x).to_bytes(2, byteorder='big', signed=False)
    def _int32(x):
        return int(x).to_bytes(4, byteorder='big', signed=True)
    def _float(x):
        return struct.pack(">f", x)

    subpixel = bool(subpixel)
    # add offset since pixel center is at (0.5,0.5) in ImageJ
    x_raw = np.asarray(x).ravel() + 0.5
    y_raw = np.asarray(y).ravel() + 0.5
    x = np.round(x_raw)
    y = np.round(y_raw)
    assert len(x) == len(y)
    top, left, bottom, right = y.min(), x.min(), y.max(), x.max() # bbox

    n_coords = len(x)
    bytes_header = 64
    bytes_total = bytes_header + n_coords*2*2 + subpixel*n_coords*2*4
    B = [0] * bytes_total
    B[ 0: 4] = map(ord,'Iout')   # magic start
    B[ 4: 6] = _int16(227)       # version
    B[ 6: 8] = _int16(0)         # roi type (0 = polygon)
    B[ 8:10] = _int16(top)       # bbox top
    B[10:12] = _int16(left)      # bbox left
    B[12:14] = _int16(bottom)    # bbox bottom
    B[14:16] = _int16(right)     # bbox right
    B[16:18] = _uint16(n_coords) # number of coordinates
    if subpixel:
        B[50:52] = _int16(128)   # subpixel resolution (option flag)
    if pos is not None:
        B[56:60] = _int32(pos)   # position (C, Z, or T)

    for i,(_x,_y) in enumerate(zip(x,y)):
        xs = bytes_header + 2*i
        ys = xs + 2*n_coords
        B[xs:xs+2] = _int16(_x - left)
        B[ys:ys+2] = _int16(_y - top)

    if subpixel:
        base1 = bytes_header + n_coords*2*2
        base2 = base1 + n_coords*4
        for i,(_x,_y) in enumerate(zip(x_raw,y_raw)):
            xs = base1 + 4*i
            ys = base2 + 4*i
            B[xs:xs+4] = _float(_x)
            B[ys:ys+4] = _float(_y)

    return bytearray(B)


def export_imagej_rois(fname, polygons, set_position=True, subpixel=True, compression=ZIP_DEFLATED):
    """ polygons assumed to be a list of arrays with shape (id,2,c) """

    if isinstance(polygons,np.ndarray):
        polygons = (polygons,)

    fname = Path(fname)
    if fname.suffix == '.zip':
        fname = fname.with_suffix('')

    with ZipFile(str(fname)+'.zip', mode='w', compression=compression) as roizip:
        for pos,polygroup in enumerate(polygons,start=1):
            for i,poly in enumerate(polygroup,start=1):
                roi = polyroi_bytearray(poly[1],poly[0], pos=(pos if set_position else None), subpixel=subpixel)
                roizip.writestr('{pos:03d}_{i:03d}.roi'.format(pos=pos,i=i), roi)


def optimize_threshold(Y, Yhat, model, nms_thresh, measure='accuracy', iou_threshs=[0.3,0.5,0.7], bracket=None, tol=1e-2, maxiter=20, verbose=1):
    """ Tune prob_thresh for provided (fixed) nms_thresh to maximize matching score (for given measure and averaged over iou_threshs). """
    np.isscalar(nms_thresh) or _raise(ValueError("nms_thresh must be a scalar"))
    iou_threshs = [iou_threshs] if np.isscalar(iou_threshs) else iou_threshs
    values = dict()

    if bracket is None:
        max_prob = max([np.max(prob) for prob, dist in Yhat])
        bracket = max_prob/2, max_prob
    # print("bracket =", bracket)

    with tqdm(total=maxiter, disable=(verbose!=1), desc="NMS threshold = %g" % nms_thresh) as progress:

        def fn(thr):
            prob_thresh = np.clip(thr, *bracket)
            value = values.get(prob_thresh)
            if value is None:
                Y_instances = [model._instances_from_prediction(y.shape, *prob_dist, prob_thresh=prob_thresh, nms_thresh=nms_thresh)[0] for y,prob_dist in zip(Y,Yhat)]
                stats = matching_dataset(Y, Y_instances, thresh=iou_threshs, show_progress=False, parallel=True)
                values[prob_thresh] = value = np.mean([s._asdict()[measure] for s in stats])
            if verbose > 1:
                print("{now}   thresh: {prob_thresh:f}   {measure}: {value:f}".format(
                    now = datetime.datetime.now().strftime('%H:%M:%S'),
                    prob_thresh = prob_thresh,
                    measure = measure,
                    value = value,
                ), flush=True)
            else:
                progress.update()
                progress.set_postfix_str("{prob_thresh:.3f} -> {value:.3f}".format(prob_thresh=prob_thresh, value=value))
                progress.refresh()
            return -value

        opt = minimize_scalar(fn, method='golden', bracket=bracket, tol=tol, options={'maxiter': maxiter})

    verbose > 1 and print('\n',opt, flush=True)
    return opt.x, -opt.fun


def _invert_dict(d):
    """ return  v-> [k_1,k_2,k_3....] for k,v in d"""
    res = defaultdict(list)
    for k,v in d.items():
        res[v].append(k)
    return res


def mask_to_categorical(y, n_classes, classes, return_cls_dict=False):
    """generates a multi-channel categorical class map

    Parameters
    ----------
    y : n-dimensional ndarray
        integer label array
    n_classes : int
        Number of different classes (without background)
    classes: dict, integer, or None
        the label to class assignment
        can be
        - dict {label -> class_id}
           the value of class_id can be
                             0   -> background class
                  1...n_classes  -> the respective object class (1 ... n_classes)
                           None  -> ignore object (prob is set to -1 for the pixels of the object, except for background class)
        - single integer value or None -> broadcast value to all labels

    Returns
    -------
    probability map of shape y.shape+(n_classes+1,) (first channel is background)

    """

    _check_label_array(y, 'y')
    if not (np.issubdtype(type(n_classes), np.integer) and n_classes>=1):
        raise ValueError(f"n_classes is '{n_classes}' but should be a positive integer")

    y_labels = np.unique(y[y>0]).tolist()

    # build dict class_id -> labels (inverse of classes)
    if np.issubdtype(type(classes), np.integer) or classes is None:
        classes = dict((k,classes) for k in y_labels)
    elif isinstance(classes, dict):
        pass
    else:
        raise ValueError("classes should be dict, single scalar, or None!")

    if not set(y_labels).issubset(set(classes.keys())):
        raise ValueError(f"all gt labels should be present in class dict provided \ngt_labels found\n{set(y_labels)}\nclass dict labels provided\n{set(classes.keys())}")

    cls_dict = _invert_dict(classes)

    # prob map
    y_mask = np.zeros(y.shape+(n_classes+1,), np.float32)

    for cls, labels in cls_dict.items():
        if cls is None:
            # prob == -1 will be used in the loss to ignore object
            y_mask[np.isin(y, labels)] = -1
        elif np.issubdtype(type(cls), np.integer) and 0 <= cls <= n_classes:
            y_mask[...,cls] = np.isin(y, labels)
        else:
            raise ValueError(f"Wrong class id '{cls}' (for n_classes={n_classes})")

    # set 0/1 background prob (unaffected by None values for class ids)
    y_mask[...,0] = (y==0)

    if return_cls_dict:
        return y_mask, cls_dict
    else:
        return y_mask


def _is_floatarray(x):
    return isinstance(x.dtype.type(0),np.floating)


def abspath(root, relpath):
    from pathlib import Path
    root = Path(root)
    if root.is_dir():
        path = root/relpath
    else:
        path = root.parent/relpath
    return str(path.absolute())



from __future__ import absolute_import, print_function

from .model2d import Config2D, StarDist2D, StarDistData2D
from .model3d import Config3D, StarDist3D, StarDistData3D

from csbdeep.utils import backend_channels_last
from csbdeep.utils.tf import keras_import
K = keras_import('backend')
if not backend_channels_last():
    raise NotImplementedError(
        "Keras is configured to use the '%s' image data format, which is currently not supported. "
        "Please change it to use 'channels_last' instead: "
        "https://keras.io/getting-started/faq/#where-is-the-keras-configuration-file-stored" % K.image_data_format()
    )
del backend_channels_last, K

from csbdeep.models import register_model, register_aliases, clear_models_and_aliases
# register pre-trained models and aliases (TODO: replace with updatable solution)
clear_models_and_aliases(StarDist2D, StarDist3D)
register_model(StarDist2D,   '2D_versatile_fluo', 'https://github.com/stardist/stardist-models/releases/download/v0.1/python_2D_versatile_fluo.zip', '8db40dacb5a1311b8d2c447ad934fb8a')
register_model(StarDist2D,   '2D_versatile_he',   'https://github.com/stardist/stardist-models/releases/download/v0.1/python_2D_versatile_he.zip', 'bf34cb3c0e5b3435971e18d66778a4ec')
register_model(StarDist2D,   '2D_paper_dsb2018',  'https://github.com/stardist/stardist-models/releases/download/v0.1/python_2D_paper_dsb2018.zip', '6287bf283f85c058ec3e7094b41039b5')
register_model(StarDist2D,   '2D_demo',           'https://github.com/stardist/stardist-models/releases/download/v0.1/python_2D_demo.zip', '31f70402f58c50dd231ec31b4375ea2c')
register_model(StarDist3D,   '3D_demo',           'https://github.com/stardist/stardist-models/releases/download/v0.1/python_3D_demo.zip', 'f481c16c1ee9f28a8dcfa1e7aae3dc83')

register_aliases(StarDist2D, '2D_paper_dsb2018',  'DSB 2018 (from StarDist 2D paper)')
register_aliases(StarDist2D, '2D_versatile_fluo', 'Versatile (fluorescent nuclei)')
register_aliases(StarDist2D, '2D_versatile_he',   'Versatile (H&E nuclei)')
del register_model, register_aliases, clear_models_and_aliases


from __future__ import absolute_import, print_function

# TODO: rethink naming for 2D/3D functions

from .geom2d import star_dist, relabel_image_stardist, ray_angles, dist_to_coord, polygons_to_label, polygons_to_label_coord
from .geom3d import star_dist3D, polyhedron_to_label, relabel_image_stardist3D, dist_to_coord3D, export_to_obj_file3D

from .geom2d import _dist_to_coord_old, _polygons_to_label_old

#, dist_to_volume, dist_to_centroid



from __future__ import print_function, unicode_literals, absolute_import, division
import numpy as np
import os

from skimage.measure import regionprops
from csbdeep.utils import _raise
from tqdm import tqdm

from ..utils import path_absolute, _normalize_grid
from ..matching import _check_label_array
# from ..lib.stardist3d import c_star_dist3d, c_polyhedron_to_label, c_dist_to_volume, c_dist_to_centroid
from ..lib.stardist3d import c_star_dist3d, c_polyhedron_to_label



def _cpp_star_dist3D(lbl, rays, grid=(1,1,1)):
    dz, dy, dx = rays.vertices.T
    grid = _normalize_grid(grid,3)

    return c_star_dist3d(lbl.astype(np.uint16, copy=False),
                         dz.astype(np.float32, copy=False),
                         dy.astype(np.float32, copy=False),
                         dx.astype(np.float32, copy=False),
                         int(len(rays)), *tuple(int(a) for a in grid))


def _py_star_dist3D(img, rays, grid=(1,1,1)):
    grid = _normalize_grid(grid,3)
    img = img.astype(np.uint16, copy=False)
    dst_shape = tuple(s // a for a, s in zip(grid, img.shape)) + (len(rays),)
    dst = np.empty(dst_shape, np.float32)

    dzs, dys, dxs = rays.vertices.T

    for i in range(dst_shape[0]):
        for j in range(dst_shape[1]):
            for k in range(dst_shape[2]):
                value = img[i * grid[0], j * grid[1], k * grid[2]]
                if value == 0:
                    dst[i, j, k] = 0
                else:

                    for n, (dz, dy, dx) in enumerate(zip(dzs, dys, dxs)):
                        x, y, z = np.float32(0), np.float32(0), np.float32(0)
                        while True:
                            x += dx
                            y += dy
                            z += dz
                            ii = int(round(i * grid[0] + z))
                            jj = int(round(j * grid[1] + y))
                            kk = int(round(k * grid[2] + x))
                            if (ii < 0 or ii >= img.shape[0] or
                                        jj < 0 or jj >= img.shape[1] or
                                        kk < 0 or kk >= img.shape[2] or
                                        value != img[ii, jj, kk]):
                                dist = np.sqrt(x * x + y * y + z * z)
                                dst[i, j, k, n] = dist
                                break

    return dst


def _ocl_star_dist3D(lbl, rays, grid=(1,1,1)):
    from gputools import OCLProgram, OCLArray, OCLImage

    grid = _normalize_grid(grid,3)

    # if not all(g==1 for g in grid):
    #     raise NotImplementedError("grid not yet implemented for OpenCL version of star_dist3D()...")

    # slicing with grid is done with tuple(slice(0, None, g) for g in grid)
    res_shape = tuple((s-1)//g+1 for s, g in zip(lbl.shape, grid))

    lbl_g = OCLImage.from_array(lbl.astype(np.uint16, copy=False))
    dist_g = OCLArray.empty(res_shape + (len(rays),), dtype=np.float32)
    rays_g = OCLArray.from_array(rays.vertices.astype(np.float32, copy=False))

    program = OCLProgram(path_absolute("kernels/stardist3d.cl"), build_options=['-D', 'N_RAYS=%d' % len(rays)])
    program.run_kernel('stardist3d', res_shape[::-1], None,
                       lbl_g, rays_g.data, dist_g.data,
                       np.int32(grid[0]),np.int32(grid[1]),np.int32(grid[2]))

    return dist_g.get()


def star_dist3D(lbl, rays, grid=(1,1,1), mode='cpp'):
    """lbl assumbed to be a label image with integer values that encode object ids. id 0 denotes background."""

    grid = _normalize_grid(grid,3)
    if mode == 'python':
        return _py_star_dist3D(lbl, rays, grid=grid)
    elif mode == 'cpp':
        return _cpp_star_dist3D(lbl, rays, grid=grid)
    elif mode == 'opencl':
        return _ocl_star_dist3D(lbl, rays, grid=grid)
    else:
        _raise(ValueError("Unknown mode %s" % mode))


def polyhedron_to_label(dist, points, rays, shape, prob=None, thr=-np.inf, labels=None, mode="full", verbose=True, overlap_label=None):
    """
    creates labeled image from stardist representations

    :param dist: array of shape (n_points,n_rays)
        the list of distances for each point and ray
    :param points: array of shape (n_points, 3)
        the list of center points
    :param rays: Rays object
        Ray object (e.g. `stardist.Rays_GoldenSpiral`) defining
        vertices and faces
    :param shape: (nz,ny,nx)
        output shape of the image
    :param prob: array of length/shape (n_points,) or None
        probability per polyhedron
    :param thr: scalar
        probability threshold (only polyhedra with prob>thr are labeled)
    :param labels: array of length/shape (n_points,) or None
        labels to use
    :param mode: str
        labeling mode, can be "full", "kernel", "hull", "bbox" or  "debug"
    :param verbose: bool
        enable to print some debug messages
    :param overlap_label: scalar or None
        if given, will label each pixel that belongs ot more than one polyhedron with that label
    :return: array of given shape
        labeled image
    """
    if len(points) == 0:
        if verbose:
            print("warning: empty list of points (returning background-only image)")
        return np.zeros(shape, np.uint16)

    dist = np.asanyarray(dist)
    points = np.asanyarray(points)

    if dist.ndim == 1:
        dist = dist.reshape(1, -1)
    if points.ndim == 1:
        points = points.reshape(1, -1)
    if labels is None:
        labels = np.arange(1, len(points) + 1)

    if np.amin(dist) <= 0:
        raise ValueError("distance array should be positive!")

    prob = np.ones(len(points)) if prob is None else np.asanyarray(prob)

    if dist.ndim != 2:
        raise ValueError("dist should be 2 dimensional but has shape %s" % str(dist.shape))

    if dist.shape[1] != len(rays):
        raise ValueError("inconsistent number of rays!")

    if len(prob) != len(points):
        raise ValueError("len(prob) != len(points)")

    if len(labels) != len(points):
        raise ValueError("len(labels) != len(points)")

    modes = {"full": 0, "kernel": 1, "hull": 2, "bbox": 3, "debug": 4}

    if not mode in modes:
        raise KeyError("Unknown render mode '%s' , allowed:  %s" % (mode, tuple(modes.keys())))

    lbl = np.zeros(shape, np.uint16)

    # filter points
    ind = np.where(prob >= thr)[0]
    if len(ind) == 0:
        if verbose:
            print("warning: no points found with probability>= {thr:.4f} (returning background-only image)".format(thr=thr))
        return lbl

    prob = prob[ind]
    points = points[ind]
    dist = dist[ind]
    labels = labels[ind]

    # sort points with decreasing probability
    ind = np.argsort(prob)[::-1]
    points = points[ind]
    dist = dist[ind]
    labels = labels[ind]

    def _prep(x, dtype):
        return np.ascontiguousarray(x.astype(dtype, copy=False))

    return c_polyhedron_to_label(_prep(dist, np.float32),
                                 _prep(points, np.float32),
                                 _prep(rays.vertices, np.float32),
                                 _prep(rays.faces, np.int32),
                                 _prep(labels, np.int32),
                                 np.int32(modes[mode]),
                                 np.int32(verbose),
                                 np.int32(overlap_label is not None),
                                 np.int32(0 if overlap_label is None else overlap_label),
                                 shape
                                 )


def relabel_image_stardist3D(lbl, rays, verbose=False, **kwargs):
    """relabel each label region in `lbl` with its star representation"""
    _check_label_array(lbl, "lbl")
    if not lbl.ndim==3:
        raise ValueError("lbl image should be 3 dimensional")

    dist_all = star_dist3D(lbl, rays, **kwargs)

    regs = regionprops(lbl)

    points = np.array(tuple(np.array(r.centroid).astype(int) for r in regs))
    labs = np.array(tuple(r.label for r in regs))
    dist = np.array(tuple(dist_all[p[0], p[1], p[2]] for p in points))
    dist = np.maximum(dist, 1e-3)

    lbl_new = polyhedron_to_label(dist, points, rays, shape=lbl.shape, labels=labs, verbose=verbose)
    return lbl_new


def dist_to_volume(dist, rays):
    """ returns areas of polyhedra
        dist.shape = (nz,ny,nx,nrays)
    """
    dist = np.asanyarray(dist)
    dist.ndim == 4 or _raise(ValueError("dist.ndim = {dist.ndim} but should be 4".format(dist = dist)))
    dist.shape[-1]== len(rays) or _raise(ValueError("dist.shape[-1] = {d} but should be {l}".format(d = dist.shape[-1], l = len(rays))))

    dist = np.ascontiguousarray(dist.astype(np.float32, copy=False))

    def _prep(x, dtype):
        return np.ascontiguousarray(x.astype(dtype, copy=False))

    return c_dist_to_volume(_prep(dist, np.float32),
                          _prep(rays.vertices, np.float32),
                          _prep(rays.faces, np.int32))


def dist_to_centroid(dist, rays, mode='absolute'):
    """ returns centroids of polyhedra

        dist.shape = (nz,ny,nx,nrays)
        mode = 'absolute' or 'relative'

    """
    dist.ndim == 4 or _raise(ValueError("dist.ndim = {dist.ndim} but should be 4".format(dist = dist)))
    dist.shape[-1]== len(rays) or _raise(ValueError("dist.shape[-1] = {d} but should be {l}".format(d = dist.shape[-1], l = len(rays))))
    dist = np.ascontiguousarray(dist.astype(np.float32, copy=False))

    mode in ('absolute', 'relative') or _raise(ValueError("mode should be either 'absolute' or 'relative'"))

    def _prep(x, dtype):
        return np.ascontiguousarray(x.astype(dtype, copy=False))

    return c_dist_to_centroid(_prep(dist, np.float32),
                          _prep(rays.vertices, np.float32),
                          _prep(rays.faces, np.int32),
                              np.int32(mode=='absolute'))



def dist_to_coord3D(dist, points, rays_vertices):
    """ converts dist/points/rays_vertices to list of coords """

    dist = np.asarray(dist)
    points = np.asarray(points)
    rays_vertices = np.asarray(rays_vertices)

    if not all((len(dist)==len(points), dist.ndim==2, points.ndim==2,
               points.shape[-1]==3, rays_vertices.shape[-1]==3, dist.shape[-1]==len(rays_vertices))):
        raise ValueError(f"Wrong shapes! dist -> (m,n) points -> (m,3) rays_vertices -> (m,)")

    # return points[:,np.newaxis]+dist[...,np.newaxis]*rays_vertices

    return points[:,np.newaxis]+dist[...,np.newaxis]*rays_vertices


def export_to_obj_file3D(polys, fname=None, scale=1, single_mesh=True, uv_map=False, name="poly"):
    """ exports 3D mesh result to obj file format """

    try:
        dist = polys["dist"]
        points = polys["points"]
        rays_vertices = polys["rays_vertices"]
        rays_faces = polys["rays_faces"]
    except KeyError as e:
        print(e)
        raise ValueError("polys should be a dict with keys 'dist', 'points', 'rays_vertices', 'rays_faces'  (such as generated by StarDist3D.predict_instances) ")

    coord = dist_to_coord3D(dist, points, rays_vertices)

    if not all((coord.ndim==3, coord.shape[-1]==3, rays_faces.shape[-1]==3)):
        raise ValueError(f"Wrong shapes! coord -> (m,n,3) rays_faces -> (k,3)")

    if np.isscalar(scale):
        scale = (scale,)*3

    scale = np.asarray(scale)
    assert len(scale)==3

    coord *= scale

    obj_str = ""
    vert_count = 0

    decimals = int(max(1,1-np.log10(np.min(scale))))


    scaled_verts = scale*rays_vertices
    scaled_verts /= np.linalg.norm(scaled_verts,axis = 1, keepdims=True)


    vertex_line = f"v {{x:.{decimals}f}} {{y:.{decimals}f}} {{z:.{decimals}f}}\n"

    rays_faces = rays_faces.copy()+1

    for i, xs in enumerate(tqdm(coord)):
        # reorder to xyz
        xs = xs[:,[2,1,0]]

        # print(xs)

        # new object
        if i==0 or not single_mesh:
            obj_str += f"o {name}_{i:d}\n"

        # vertex coords
        for x,y,z in xs:
            obj_str += vertex_line.format(x=x,y=y,z=z)

        if uv_map:
            # UV coords
            for vz,vy,vx in scaled_verts:
                u = 1-(.5 + .5*np.arctan2(vz,vx)/np.pi)
                v = 1-(.5 - np.arcsin(vy)/np.pi)
                obj_str +=  f"vt {u:.4f} {v:.4f}\n"

        # face indices
        for face in rays_faces:
            obj_str += f"f {face[0]}/{face[0]} {face[1]}/{face[1]} {face[2]}/{face[2]}\n"

        rays_faces += len(xs)

    if fname is not None:
        with open(fname,"w") as f:
            f.write(obj_str)

    return obj_str





from __future__ import absolute_import, print_function

import warnings
def format_warning(message, category, filename, lineno, line=''):
    import pathlib
    return f"{pathlib.Path(filename).name} ({lineno}): {message}\n"
warnings.formatwarning = format_warning
del warnings

from .version import __version__

# TODO: which functions to expose here? all?
from .nms import non_maximum_suppression, non_maximum_suppression_3d, non_maximum_suppression_3d_sparse
from .utils import edt_prob, fill_label_holes, sample_points, calculate_extents, export_imagej_rois, gputools_available
from .geometry import star_dist,   polygons_to_label,   relabel_image_stardist, ray_angles, dist_to_coord
from .geometry import star_dist3D, polyhedron_to_label, relabel_image_stardist3D
from .plot.plot import random_label_cmap, draw_polygons, _draw_polygons
from .plot.render import render_label, render_label_pred
from .rays3d import rays_from_json, Rays_Cartesian, Rays_SubDivide, Rays_Tetra, Rays_Octo, Rays_GoldenSpiral, Rays_Explicit
from .sample_patches import sample_patches
from .bioimageio_utils import export_bioimageio, import_bioimageio

def _py_deprecation(ver_python=(3,6), ver_stardist='0.9.0'):
     import sys
     from distutils.version import LooseVersion
     if sys.version_info[:2] == ver_python and LooseVersion(__version__) < LooseVersion(ver_stardist):
         print(f"You are using Python {ver_python[0]}.{ver_python[1]}, which will no longer be supported in StarDist {ver_stardist}.\n"
               f"→ Please upgrade to Python {ver_python[0]}.{ver_python[1]+1} or later.", file=sys.stderr, flush=True)
_py_deprecation()
del _py_deprecation


from pathlib import Path
from pkg_resources import get_distribution
from zipfile import ZipFile
import numpy as np
import tempfile
from distutils.version import LooseVersion
from csbdeep.utils import axes_check_and_normalize, normalize, _raise


DEEPIMAGEJ_MACRO = \
"""
//*******************************************************************
// Date: July-2021
// Credits: StarDist, DeepImageJ
// URL:
//      https://github.com/stardist/stardist
//      https://deepimagej.github.io/deepimagej
// This macro was adapted from
// https://github.com/deepimagej/imagej-macros/blob/648caa867f6ccb459649d4d3799efa1e2e0c5204/StarDist2D_Post-processing.ijm
// Please cite the respective contributions when using this code.
//*******************************************************************
//  Macro to run StarDist postprocessing on 2D images.
//  StarDist and deepImageJ plugins need to be installed.
//  The macro assumes that the image to process is a stack in which
//  the first channel corresponds to the object probability map
//  and the remaining channels are the radial distances from each
//  pixel to the object boundary.
//*******************************************************************

// Get the name of the image to call it
getDimensions(width, height, channels, slices, frames);
name=getTitle();

probThresh={probThresh};
nmsThresh={nmsThresh};

// Isolate the detection probability scores
run("Make Substack...", "channels=1");
rename("scores");

// Isolate the oriented distances
run("Fire");
selectWindow(name);
run("Delete Slice", "delete=channel");
selectWindow(name);
run("Properties...", "channels=" + maxOf(channels, slices) - 1 + " slices=1 frames=1 pixel_width=1.0000 pixel_height=1.0000 voxel_depth=1.0000");
rename("distances");
run("royal");

// Run StarDist plugin
run("Command From Macro", "command=[de.csbdresden.stardist.StarDist2DNMS], args=['prob':'scores', 'dist':'distances', 'probThresh':'" + probThresh + "', 'nmsThresh':'" + nmsThresh + "', 'outputType':'Both', 'excludeBoundary':'2', 'roiPosition':'Stack', 'verbose':'false'], process=[false]");
"""


def _import(error=True):
    try:
        from importlib_metadata import metadata
        from bioimageio.core.build_spec import build_model # type: ignore
        import xarray as xr
        import bioimageio.core # type: ignore
    except ImportError:
        if error:
            raise RuntimeError(
                "Required libraries are missing for bioimage.io model export.\n"
                "Please install StarDist as follows: pip install 'stardist[bioimageio]'\n"
                "(You do not need to uninstall StarDist first.)"
            )
        else:
            return None
    return metadata, build_model, bioimageio.core, xr


def _create_stardist_dependencies(outdir):
    from ruamel.yaml import YAML
    from tensorflow import __version__ as tf_version
    from . import __version__ as stardist_version
    pkg_info = get_distribution("stardist")
    # dependencies that start with the name "bioimageio" will be added as conda dependencies
    reqs_conda = [str(req) for req in pkg_info.requires(extras=['bioimageio']) if str(req).startswith('bioimageio')]
    # only stardist and tensorflow as pip dependencies
    tf_major, tf_minor = LooseVersion(tf_version).version[:2]
    reqs_pip = (f"stardist>={stardist_version}", f"tensorflow>={tf_major}.{tf_minor},<{tf_major+1}")
    # conda environment
    env = dict(
        name = 'stardist',
        channels = ['defaults', 'conda-forge'],
        dependencies = [
            ('python>=3.7,<3.8' if tf_major == 1 else 'python>=3.7'),
            *reqs_conda,
            'pip', {'pip': reqs_pip},
        ],
    )
    yaml = YAML(typ='safe')
    path = outdir / "environment.yaml"
    with open(path, "w") as f:
        yaml.dump(env, f)
    return f"conda:{path}"


def _create_stardist_doc(outdir):
    doc_path = outdir / "README.md"
    text = (
        "# StarDist Model\n"
        "This is a model for object detection with star-convex shapes.\n"
        "Please see the [StarDist repository](https://github.com/stardist/stardist) for details."
    )
    with open(doc_path, "w") as f:
        f.write(text)
    return doc_path


def _get_stardist_metadata(outdir, model):
    metadata, *_ = _import()
    package_data = metadata("stardist")
    doi_2d = "https://doi.org/10.1007/978-3-030-00934-2_30"
    doi_3d = "https://doi.org/10.1109/WACV45572.2020.9093435"
    authors = {
        'Martin Weigert': dict(name='Martin Weigert', github_user='maweigert'),
        'Uwe Schmidt': dict(name='Uwe Schmidt', github_user='uschmidt83'),
    }
    data = dict(
        description=package_data["Summary"],
        authors=list(authors.get(name.strip(),dict(name=name.strip())) for name in package_data["Author"].split(",")),
        git_repo=package_data["Home-Page"],
        license=package_data["License"],
        dependencies=_create_stardist_dependencies(outdir),
        cite=[{"text": "Cell Detection with Star-Convex Polygons", "doi": doi_2d},
              {"text": "Star-convex Polyhedra for 3D Object Detection and Segmentation in Microscopy", "doi": doi_3d}],
        tags=[
            'fluorescence-light-microscopy', 'whole-slide-imaging', 'other', # modality
            f'{model.config.n_dim}d', # dims
            'cells', 'nuclei', # content
            'tensorflow', # framework
            'fiji', # software
            'unet', # network
            'instance-segmentation', 'object-detection', # task
            'stardist',
        ],
        covers=["https://raw.githubusercontent.com/stardist/stardist/master/images/stardist_logo.jpg"],
        documentation=_create_stardist_doc(outdir),
    )
    return data


def _predict_tf(model_path, test_input):
    import tensorflow as tf
    from csbdeep.utils.tf import IS_TF_1
    # need to unzip the model assets
    model_assets = model_path.parent / "tf_model"
    with ZipFile(model_path, "r") as f:
        f.extractall(model_assets)
    if IS_TF_1:
        # make a new graph, i.e. don't use the global default graph
        with tf.Graph().as_default():
            with tf.Session() as sess:
                tf_model = tf.saved_model.load_v2(str(model_assets))
                x = tf.convert_to_tensor(test_input, dtype=tf.float32)
                model = tf_model.signatures["serving_default"]
                y = model(x)
                sess.run(tf.global_variables_initializer())
                output = sess.run(y["output"])
    else:
        tf_model = tf.saved_model.load(str(model_assets))
        x = tf.convert_to_tensor(test_input, dtype=tf.float32)
        model = tf_model.signatures["serving_default"]
        y = model(x)
        output = y["output"].numpy()
    return output


def _get_weights_and_model_metadata(outdir, model, test_input, test_input_axes, test_input_norm_axes, mode, min_percentile, max_percentile):

    # get the path to the exported model assets (saved in outdir)
    if mode == "keras_hdf5":
        raise NotImplementedError("Export to keras format is not supported yet")
    elif mode == "tensorflow_saved_model_bundle":
        assets_uri = outdir / "TF_SavedModel.zip"
        model_csbdeep = model.export_TF(assets_uri, single_output=True, upsample_grid=True)
    else:
        raise ValueError(f"Unsupported mode: {mode}")

    # to force "inputs.data_type: float32" in the spec (bonus: disables normalization warning in model._predict_setup)
    test_input = test_input.astype(np.float32)

    # convert test_input to axes_net semantics and shape, also resize if necessary (to adhere to axes_net_div_by)
    test_input, axes_img, axes_net, axes_net_div_by, *_ = model._predict_setup(
        img=test_input,
        axes=test_input_axes,
        normalizer=None,
        n_tiles=None,
        show_tile_progress=False,
        predict_kwargs={},
    )

    # normalization axes string and numeric indices
    axes_norm = set(axes_net).intersection(set(axes_check_and_normalize(test_input_norm_axes, disallowed='S')))
    axes_norm = "".join(a for a in axes_net if a in axes_norm)  # preserve order of axes_net
    axes_norm_num = tuple(axes_net.index(a) for a in axes_norm)

    # normalize input image
    test_input_norm = normalize(test_input, pmin=min_percentile, pmax=max_percentile, axis=axes_norm_num)

    net_axes_in = axes_net.lower()
    net_axes_out = axes_check_and_normalize(model._axes_out).lower()
    ndim_tensor = len(net_axes_out) + 1

    input_min_shape = list(axes_net_div_by)
    input_min_shape[axes_net.index('C')] = model.config.n_channel_in
    input_step = list(axes_net_div_by)
    input_step[axes_net.index('C')] = 0

    # add the batch axis to shape and step
    input_min_shape = [1] + input_min_shape
    input_step = [0] + input_step

    # the axes strings in bioimageio convention
    input_axes = "b" + net_axes_in.lower()
    output_axes = "b" + net_axes_out.lower()

    if mode == "keras_hdf5":
        output_names = ("prob", "dist") + (("class_prob",) if model._is_multiclass() else ())
        output_n_channels = (1, model.config.n_rays,) + ((1,) if model._is_multiclass() else ())
        # the output shape is computed from the input shape using
        # output_shape[i] = output_scale[i] * input_shape[i] + 2 * output_offset[i]
        output_scale = [1]+list(1/g for g in model.config.grid) + [0]
        output_offset = [0]*(ndim_tensor)

    elif mode == "tensorflow_saved_model_bundle":
        if model._is_multiclass():
            raise NotImplementedError("Tensorflow SavedModel not supported for multiclass models yet")
        # regarding input/output names: https://github.com/CSBDeep/CSBDeep/blob/b0d2f5f344ebe65a9b4c3007f4567fe74268c813/csbdeep/utils/tf.py#L193-L194
        input_names = ["input"]
        output_names = ["output"]
        output_n_channels = (1 + model.config.n_rays,)
        # the output shape is computed from the input shape using
        # output_shape[i] = output_scale[i] * input_shape[i] + 2 * output_offset[i]
        # same shape as input except for the channel dimension
        output_scale = [1]*(ndim_tensor)
        output_scale[output_axes.index("c")] = 0
        # no offset, except for the input axes, where it is output channel / 2
        output_offset = [0.0]*(ndim_tensor)
        output_offset[output_axes.index("c")] = output_n_channels[0] / 2.0

    assert all(s in (0, 1) for s in output_scale), "halo computation assumption violated"
    halo = model._axes_tile_overlap(output_axes.replace('b', 's'))
    halo = [int(np.ceil(v/8)*8) for v in halo]  # optional: round up to be divisible by 8

    # the output shape needs to be valid after cropping the halo, so we add the halo to the input min shape
    input_min_shape = [ms + 2 * ha for ms, ha in zip(input_min_shape, halo)]

    # make sure the input min shape is still divisible by the min axis divisor
    input_min_shape = input_min_shape[:1] + [ms + (-ms % div_by) for ms, div_by in zip(input_min_shape[1:], axes_net_div_by)]
    assert all(ms % div_by == 0 for ms, div_by in zip(input_min_shape[1:], axes_net_div_by))

    metadata, *_ = _import()
    package_data = metadata("stardist")
    is_2D = model.config.n_dim == 2

    weights_file = outdir / "stardist_weights.h5"
    model.keras_model.save_weights(str(weights_file))

    config = dict(
        stardist=dict(
            python_version=package_data["Version"],
            thresholds=dict(model.thresholds._asdict()),
            weights=weights_file.name,
            config=vars(model.config),
        )
    )

    if is_2D:
        macro_file = outdir / "stardist_postprocessing.ijm"
        with open(str(macro_file), 'w', encoding='utf-8') as f:
            f.write(DEEPIMAGEJ_MACRO.format(probThresh=model.thresholds.prob, nmsThresh=model.thresholds.nms))
        config['stardist'].update(postprocessing_macro=macro_file.name)

    n_inputs = len(input_names)
    assert n_inputs == 1
    input_config = dict(
        input_names=input_names,
        input_min_shape=[input_min_shape],
        input_step=[input_step],
        input_axes=[input_axes],
        input_data_range=[["-inf", "inf"]],
        preprocessing=[[dict(
            name="scale_range",
            kwargs=dict(
                mode="per_sample",
                axes=axes_norm.lower(),
                min_percentile=min_percentile,
                max_percentile=max_percentile,
            ))]]
        )

    n_outputs = len(output_names)
    output_config = dict(
        output_names=output_names,
        output_data_range=[["-inf", "inf"]] * n_outputs,
        output_axes=[output_axes] * n_outputs,
        output_reference=[input_names[0]] * n_outputs,
        output_scale=[output_scale] * n_outputs,
        output_offset=[output_offset] * n_outputs,
        halo=[halo] * n_outputs
    )

    in_path = outdir / "test_input.npy"
    np.save(in_path, test_input[np.newaxis])

    if mode == "tensorflow_saved_model_bundle":
        test_outputs = _predict_tf(assets_uri, test_input_norm[np.newaxis])
    else:
        test_outputs = model.predict(test_input_norm)

    # out_paths = []
    # for i, out in enumerate(test_outputs):
    #     p = outdir / f"test_output{i}.npy"
    #     np.save(p, out)
    #     out_paths.append(p)
    assert n_outputs == 1
    out_paths = [outdir / "test_output.npy"]
    np.save(out_paths[0], test_outputs)

    from tensorflow import __version__ as tf_version
    data = dict(weight_uri=assets_uri, test_inputs=[in_path], test_outputs=out_paths,
                config=config, tensorflow_version=tf_version)
    data.update(input_config)
    data.update(output_config)
    _files = [str(weights_file)]
    if is_2D:
        _files.append(str(macro_file))
    data.update(attachments=dict(files=_files))

    return data


def export_bioimageio(
    model,
    outpath,
    test_input,
    test_input_axes=None,
    test_input_norm_axes='ZYX',
    name=None,
    mode="tensorflow_saved_model_bundle",
    min_percentile=1.0,
    max_percentile=99.8,
    overwrite_spec_kwargs=None,
):
    """Export stardist model into bioimage.io format, https://github.com/bioimage-io/spec-bioimage-io.

    Parameters
    ----------
    model: StarDist2D, StarDist3D
        the model to convert
    outpath: str, Path
        where to save the model
    test_input: np.ndarray
        input image for generating test data
    test_input_axes: str or None
        the axes of the test input, for example 'YX' for a 2d image or 'ZYX' for a 3d volume
        using None assumes that axes of test_input are the same as those of model
    test_input_norm_axes: str
        the axes of the test input which will be jointly normalized, for example 'ZYX' for all spatial dimensions ('Z' ignored for 2D input)
        use 'ZYXC' to also jointly normalize channels (e.g. for RGB input images)
    name: str
        the name of this model (default: None)
        if None, uses the (folder) name of the model (i.e. `model.name`)
    mode: str
        the export type for this model (default: "tensorflow_saved_model_bundle")
    min_percentile: float
        min percentile to be used for image normalization (default: 1.0)
    max_percentile: float
        max percentile to be used for image normalization (default: 99.8)
    overwrite_spec_kwargs: dict or None
        spec keywords that should be overloaded (default: None)
    """
    _, build_model, *_ = _import()
    from .models import StarDist2D, StarDist3D
    isinstance(model, (StarDist2D, StarDist3D)) or _raise(ValueError("not a valid model"))
    0 <= min_percentile < max_percentile <= 100 or _raise(ValueError("invalid percentile values"))

    if name is None:
        name = model.name
    name = str(name)

    outpath = Path(outpath)
    if outpath.suffix == "":
        outdir = outpath
        zip_path = outdir / f"{name}.zip"
    elif outpath.suffix == ".zip":
        outdir = outpath.parent
        zip_path = outpath
    else:
        raise ValueError(f"outpath has to be a folder or zip file, got {outpath}")
    outdir.mkdir(exist_ok=True, parents=True)

    with tempfile.TemporaryDirectory() as _tmp_dir:
        tmp_dir = Path(_tmp_dir)
        kwargs = _get_stardist_metadata(tmp_dir, model)
        model_kwargs = _get_weights_and_model_metadata(tmp_dir, model, test_input, test_input_axes, test_input_norm_axes, mode,
                                                       min_percentile=min_percentile, max_percentile=max_percentile)
        kwargs.update(model_kwargs)
        if overwrite_spec_kwargs is not None:
            kwargs.update(overwrite_spec_kwargs)

        build_model(name=name, output_path=zip_path, add_deepimagej_config=(model.config.n_dim==2), root=tmp_dir, **kwargs)
        print(f"\nbioimage.io model with name '{name}' exported to '{zip_path}'")


def import_bioimageio(source, outpath):
    """Import stardist model from bioimage.io format, https://github.com/bioimage-io/spec-bioimage-io.

    Load a model in bioimage.io format from the given `source` (e.g. path to zip file, URL)
    and convert it to a regular stardist model, which will be saved in the folder `outpath`.

    Parameters
    ----------
    source: str, Path
        bioimage.io resource (e.g. path, URL)
    outpath: str, Path
        folder to save the stardist model (must not exist previously)

    Returns
    -------
    StarDist2D or StarDist3D
        stardist model loaded from `outpath`

    """
    import shutil, uuid
    from csbdeep.utils import save_json
    from .models import StarDist2D, StarDist3D
    *_, bioimageio_core, _ = _import()

    outpath = Path(outpath)
    not outpath.exists() or _raise(FileExistsError(f"'{outpath}' already exists"))

    with tempfile.TemporaryDirectory() as _tmp_dir:
        tmp_dir = Path(_tmp_dir)
        # download the full model content to a temporary folder
        zip_path = tmp_dir / f"{str(uuid.uuid4())}.zip"
        bioimageio_core.export_resource_package(source, output_path=zip_path)
        with ZipFile(zip_path, "r") as zip_ref:
            zip_ref.extractall(tmp_dir)
        zip_path.unlink()
        rdf_path = tmp_dir / "rdf.yaml"
        biomodel = bioimageio_core.load_resource_description(rdf_path)

        # read the stardist specific content
        'stardist' in biomodel.config or _raise(RuntimeError("bioimage.io model not compatible"))
        config = biomodel.config['stardist']['config']
        thresholds = biomodel.config['stardist']['thresholds']
        weights = biomodel.config['stardist']['weights']

        # make sure that the keras weights are in the attachments
        weights_file = None
        for f in biomodel.attachments.files:
            if f.name == weights and f.exists():
                weights_file = f
                break
        weights_file is not None or _raise(FileNotFoundError(f"couldn't find weights file '{weights}'"))

        # save the config and threshold to json, and weights to hdf5 to enable loading as stardist model
        # copy bioimageio files to separate sub-folder
        outpath.mkdir(parents=True)
        save_json(config, str(outpath / 'config.json'))
        save_json(thresholds, str(outpath / 'thresholds.json'))
        shutil.copy(str(weights_file), str(outpath / "weights_bioimageio.h5"))
        shutil.copytree(str(tmp_dir), str(outpath / "bioimageio"))

    model_class = (StarDist2D if config['n_dim'] == 2 else StarDist3D)
    model = model_class(None, outpath.name, basedir=str(outpath.parent))

    return model



from __future__ import print_function, unicode_literals, absolute_import, division
import numpy as np
from csbdeep.utils import normalize
from ..matching import matching
from .plot import random_label_cmap


def _single_color_integer_cmap(color = (.3,.4,.5)):
    from matplotlib.colors import Colormap
    
    assert len(color) in (3,4)
    
    class BinaryMap(Colormap):
        def __init__(self, color):
            self.color = np.array(color)
            if len(self.color)==3:
                self.color = np.concatenate([self.color,[1]])
        def __call__(self, X, alpha=None, bytes=False):
            res = np.zeros(X.shape+(4,), np.float32)
            res[...,-1] = self.color[-1]
            res[X>0] = np.expand_dims(self.color,0)
            if bytes:
                return np.clip(256*res,0,255).astype(np.uint8)
            else:
                return res
    return BinaryMap(color)
                       



def render_label(lbl, img = None, cmap = None, cmap_img = "gray", alpha = 0.5, alpha_boundary = None, normalize_img = True):
    """Renders a label image and optionally overlays it with another image. Used for generating simple output images to asses the label quality

    Parameters
    ----------
    lbl: np.ndarray of dtype np.uint16
        The 2D label image 
    img: np.ndarray 
        The array to overlay the label image with (optional)
    cmap: string, tuple, or callable
        The label colormap. If given as rgb(a)  only a single color is used, if None uses a random colormap 
    cmap_img: string or callable
        The colormap of img (optional)
    alpha: float 
        The alpha value of the overlay. Set alpha=1 to get fully opaque labels
    alpha_boundary: float
        The alpha value of the boundary (if None, use the same as for labels, i.e. no boundaries are visible)
    normalize_img: bool
        If True, normalizes the img (if given)

    Returns
    -------
    img: np.ndarray
        the (m,n,4) RGBA image of the rendered label 

    Example
    -------

    from scipy.ndimage import label, zoom      
    img = zoom(np.random.uniform(0,1,(16,16)),(8,8),order=3)            
    lbl,_ = label(img>.8)
    u1 = render_label(lbl, img = img, alpha = .7)
    u2 = render_label(lbl, img = img, alpha = 0, alpha_boundary =.8)
    plt.subplot(1,2,1);plt.imshow(u1)
    plt.subplot(1,2,2);plt.imshow(u2)

    """
    from skimage.segmentation import find_boundaries
    from matplotlib import cm
    
    alpha = np.clip(alpha, 0, 1)

    if alpha_boundary is None:
        alpha_boundary = alpha
        
    if cmap is None:
        cmap = random_label_cmap()
    elif isinstance(cmap, tuple):
        cmap = _single_color_integer_cmap(cmap)
    else:
        pass
        
    cmap = cm.get_cmap(cmap) if isinstance(cmap, str) else cmap
    cmap_img = cm.get_cmap(cmap_img) if isinstance(cmap_img, str) else cmap_img

    # render image if given
    if img is None:
        im_img = np.zeros(lbl.shape+(4,),np.float32)
        im_img[...,-1] = 1
        
    else:
        assert lbl.shape[:2] == img.shape[:2]
        img = normalize(img) if normalize_img else img
        if img.ndim==2:
            im_img = cmap_img(img)
        elif img.ndim==3:
            im_img = img[...,:4]
            if img.shape[-1]<4:
                im_img = np.concatenate([img, np.ones(img.shape[:2]+(4-img.shape[-1],))], axis = -1)
        else:
            raise ValueError("img should be 2 or 3 dimensional")
            
                
            
    # render label
    im_lbl = cmap(lbl)

    mask_lbl = lbl>0
    mask_bound = np.bitwise_and(mask_lbl,find_boundaries(lbl, mode = "thick"))
    
    # blend
    im = im_img.copy()
    
    im[mask_lbl] = alpha*im_lbl[mask_lbl]+(1-alpha)*im_img[mask_lbl]
    im[mask_bound] = alpha_boundary*im_lbl[mask_bound]+(1-alpha_boundary)*im_img[mask_bound]
        
    return im


def random_hls(n=2**16, h0 = .33, l0 = (.8,1), s0 = (.5,.8)  ):
    """
    h0 = 0 -> red
    h0 = 0.33 -> green
    h0 = 0.66 -> blue
    h0 = 0.833 -> magenta
    """
    _f = lambda x: (x,)*2 if np.isscalar(x) else tuple(x)
    h0,s0,l0 = map(_f,(h0,s0,l0))

    h = np.random.uniform(*h0,n)
    s = np.random.uniform(*s0,n)
    l = np.random.uniform(*l0,n)

    return h,l,s

def cmap_from_hls(h,l,s):
    import matplotlib
    import colorsys
    h = h%1
    l = np.clip(l,0,1)
    s = np.clip(s,0,1)

    cols = np.stack([colorsys.hls_to_rgb(_h,_l,_s) for _h,_l,_s in zip(h,l,s)],axis=0)
    cols[0] = 0
    return matplotlib.colors.ListedColormap(cols)

def match_labels(y0,y):
    """match labels from y to y0"""
    res = matching(y0,y,report_matches=True, thresh =.1)
    if len(res.matched_pairs)==0:
        print("no matching found")
        return y
    
    ind_matched0, ind_matched = tuple(zip(*res.matched_pairs))
    
    ind_unmatched = (set(np.unique(y))-{0}) - set(ind_matched)

    leftover_labels = set(np.arange(1,np.max(ind_matched0))) - set(ind_matched0)

    leftover_labels = leftover_labels.union(set(np.max(ind_matched0)+1+np.arange(len(ind_unmatched)-len(leftover_labels))))

    assert len(leftover_labels)>= len(ind_unmatched)

    
    print(f"matched: {len(ind_matched)} unmatched: {len(ind_unmatched)}")
    u = np.zeros_like(y)
    for ind0,ind in zip(ind_matched0, ind_matched):
        u[y==ind] = ind0

    for ind,ind2 in zip(ind_unmatched, leftover_labels):
        u[y==ind] = ind2
    
    return u



def render_label_pred(y_true, y_pred,
                      img = None, cmap_img = "gray", normalize_img = True, 
                      tp_alpha = .6, fp_alpha = .6, fn_alpha = .6,
                      matching_kwargs = dict(thresh=0.5)):
    """Renders an image that shows segmentation errors between y_true and y_pred 

    Correctly matched objects (TP) are colored green (with transparency tp_alpha)
    Erronously detected objects (FP) are colored red (with transparency fp_alpha)
    MIssing GT objects (FN) are colored blue (with transparency fn_alpha)

    Parameters
    ----------
    y_true: np.ndarray of dtype np.uint16
        The 2D GT label image 
    y_pred: np.ndarray of dtype np.uint16
        The 2D prediction label image 
    img: np.ndarray 
        The array to overlay the label image with (optional)
    cmap_img: string or callable
        The colormap of img (optional)
    tp_alpha: float 
        The alpha value of the TP 
    fp_alpha: float 
        The alpha value of the FP 
    fn_alpha: float 
        The alpha value of the FN 
    matching_kwargs: dict
        The parameters of stardist.matching.matching that are used to compute the TP/FP/FN
    normalize_img: bool
        If True, normalizes the img (if given)

    Returns
    -------
    img: np.ndarray
        the (m,n,4) RGBA image of the rendered image

    Example
    ------- 

    from scipy.ndimage import label, zoom      
    img = zoom(np.random.uniform(0,1,(16,16)),(8,8),order=3)
    y_true = label(img>.9)[0]
    y_pred = label(img>.02)[0]
    plt.imshow(render_label_pred(y_true, y_pred, img=img))

    """
    
    from matplotlib import cm

    assert y_true.shape == y_pred.shape
    
    matching_kwargs["report_matches"] = True
    res = matching(y_true, y_pred, **matching_kwargs)

    all_true = set(np.unique(y_true))-{0}
    all_pred = set(np.unique(y_pred))-{0}

    pairs = np.array(res.matched_pairs)
    scores = np.array(res.matched_scores)
    ind_tp_pairs = np.where(scores>=matching_kwargs["thresh"])[0]
    tp_true, tp_pred = tuple(zip(*pairs[ind_tp_pairs]))
    tp = tp_pred
    fn = all_true.difference(tp_true)
    fp = all_pred.difference(tp_pred)

    assert res.tp == len(tp)
    assert res.fp == len(fp)
    assert res.fn == len(fn)

    mask_tp = np.isin(y_pred, tuple(tp))
    mask_fn = np.isin(y_true, tuple(fn))
    mask_fp = np.isin(y_pred, tuple(fp))

    def gen_maps(n, h0,l0,s0):
        h,l,s = random_hls(n,h0,l0,s0)
        return cmap_from_hls(h,l,s), cmap_from_hls(h,1.4*l,s)

    n0 = np.max(y_pred)+1

    # green
    cmap_tp, cmap_border_tp = gen_maps(n0, h0 = (.25,.35) , l0 = (.4,.6), s0 = (.5,.7))
    # red
    cmap_fp, cmap_border_fp = gen_maps(n0, h0 = (0,.1) , l0 = (.4,.6), s0 = (.5,.7))
    # blue
    cmap_fn, cmap_border_fn = gen_maps(n0, h0 = (.6,.7) , l0 = (.4,.6), s0 = (.5,.7))

    im_tp = cmap_tp(y_pred)
    im_fp = cmap_fp(y_pred)
    im_fn = cmap_fn(y_true)

    # render image if given
    if img is None:
        im_img = np.zeros(y_true.shape+(4,),np.float32)
        im_img[...,-1] = 1
    else:
        assert y_true.shape[:2] == img.shape[:2]
        img = normalize(img) if normalize_img else img
        cmap_img = cm.get_cmap(cmap_img) if isinstance(cmap_img, str) else cmap_img
        if img.ndim==2:
            im_img = cmap_img(img)
        elif img.ndim==3:
            im_img = img[...,:4]
            if img.shape[-1]<4:
                im_img = np.concatenate([img, np.ones(img.shape[:2]+(4-img.shape[-1],))], axis = -1)
        else:
            raise ValueError("img should be 2 or 3 dimensional")
        
        
    # blend
    im = im_img.copy()
    
    im[mask_tp] = tp_alpha*im_tp[mask_tp]+(1-tp_alpha)*im_img[mask_tp]
    im[mask_fp] = fp_alpha*im_fp[mask_fp]+(1-fp_alpha)*im_img[mask_fp]
    im[mask_fn] = fn_alpha*im_fn[mask_fn]+(1-fn_alpha)*im_img[mask_fn]
    return im
